use crate::{
    data::varint::{read_varint, size_varint},
    zigzag::Zigzag,
    ProtocolError,
};
use tokio::io::AsyncReadExt;
use uuid::Uuid;

/// Packet data reader trait
pub trait DataReader {
    /// Read bytes
    fn read_bytes(&mut self, size: usize) -> impl Future<Output = Result<Vec<u8>, ProtocolError>>;

    /// Read byte
    fn read_byte(&mut self) -> impl Future<Output = Result<u8, ProtocolError>> {
        async {Ok(self.read_bytes(1).await?[0])}
    }
    /// Read String
    fn read_string(&mut self) -> impl Future<Output = Result<String, ProtocolError>> {
        async {
            let size = self.read_usize_varint().await?;
            match String::from_utf8(self.read_bytes(size).await?) {
                Ok(i) => Ok(i),
                Err(_) => Err(ProtocolError::StringParseError),
            }
        }
    }
    /// Read Unsigned Short as u16
    fn read_unsigned_short(&mut self) -> impl Future<Output = Result<u16, ProtocolError>> {
        async {match self.read_bytes(2).await?.try_into() {
            Ok(i) => Ok(u16::from_be_bytes(i)),
            Err(_) => Err(ProtocolError::ReadError),
        }}
    }
    /// Read Boolean
    fn read_boolean(&mut self) -> impl Future<Output = Result<bool, ProtocolError>> {
        async {Ok(self.read_byte().await? == 0x01)}
    }
    /// Read Short as i16
    fn read_short(&mut self) -> impl Future<Output = Result<i16, ProtocolError>> {
        async {match self.read_bytes(2).await?.try_into() {
            Ok(i) => Ok(i16::from_be_bytes(i)),
            Err(_) => Err(ProtocolError::ReadError),
        }}
    }
    /// Read Long as i64
    fn read_long(&mut self) -> impl Future<Output = Result<i64, ProtocolError>> {
        async {match self.read_bytes(8).await?.try_into() {
            Ok(i) => Ok(i64::from_be_bytes(i)),
            Err(_) => Err(ProtocolError::ReadError),
        }}
    }
    /// Read Float as f32
    fn read_float(&mut self) -> impl Future<Output = Result<f32, ProtocolError>> {
        async {match self.read_bytes(4).await?.try_into() {
            Ok(i) => Ok(f32::from_be_bytes(i)),
            Err(_) => Err(ProtocolError::ReadError),
        }}
    }
    /// Read Double as f64
    fn read_double(&mut self) -> impl Future<Output = Result<f64, ProtocolError>> {
        async {match self.read_bytes(8).await?.try_into() {
            Ok(i) => Ok(f64::from_be_bytes(i)),
            Err(_) => Err(ProtocolError::ReadError),
        }}
    }
    /// Read Int as i32
    fn read_int(&mut self) -> impl Future<Output = Result<i32, ProtocolError>> {
        async {match self.read_bytes(4).await?.try_into() {
            Ok(i) => Ok(i32::from_be_bytes(i)),
            Err(_) => Err(ProtocolError::ReadError),
        }}
    }
    /// Read UUID
    fn read_uuid(&mut self) -> impl Future<Output = Result<Uuid, ProtocolError>> {
        async {match self.read_bytes(16).await?.try_into() {
            Ok(i) => Ok(Uuid::from_bytes(i)),
            Err(_) => Err(ProtocolError::ReadError),
        }}
    }

    /// Read VarInt as usize with size in bytes (varint, size)
    fn read_usize_varint_size(&mut self) -> impl Future<Output = Result<(usize, usize), ProtocolError>> {
        async {size_varint!(usize, self)}
    }
    /// Read VarInt as u8 with size in bytes (varint, size)
    fn read_u8_varint_size(&mut self) -> impl Future<Output = Result<(u8, usize), ProtocolError>> {
        async {size_varint!(u8, self)}
    }
    /// Read VarInt as u16 with size in bytes (varint, size)
    fn read_u16_varint_size(&mut self) -> impl Future<Output = Result<(u16, usize), ProtocolError>> {
        async {size_varint!(u16, self)}
    }
    /// Read VarInt as u32 with size in bytes (varint, size)
    fn read_u32_varint_size(&mut self) -> impl Future<Output = Result<(u32, usize), ProtocolError>> {
        async {size_varint!(u32, self)}
    }
    /// Read VarInt as u64 with size in bytes (varint, size)
    fn read_u64_varint_size(&mut self) -> impl Future<Output = Result<(u64, usize), ProtocolError>> {
        async {size_varint!(u64, self)}
    }
    /// Read VarInt as u128 with size in bytes (varint, size)
    fn read_u128_varint_size(&mut self) -> impl Future<Output = Result<(u128, usize), ProtocolError>> {
        async {size_varint!(u128, self)}
    }

    /// Read VarInt as isize with size in bytes (varint, size)
    fn read_isize_varint_size(&mut self) -> impl Future<Output = Result<(isize, usize), ProtocolError>> {
        async {Ok({
            let i = self.read_usize_varint_size().await?;
            (i.0.zigzag(), i.1)
        })}
    }
    /// Read VarInt as i8 with size in bytes (varint, size)
    fn read_i8_varint_size(&mut self) -> impl Future<Output = Result<(i8, usize), ProtocolError>> {
        async {Ok({
            let i = self.read_u8_varint_size().await?;
            (i.0.zigzag(), i.1)
        })}
    }
    /// Read VarInt as i16 with size in bytes (varint, size)
    fn read_i16_varint_size(&mut self) -> impl Future<Output = Result<(i16, usize), ProtocolError>> {
        async {Ok({
            let i = self.read_u16_varint_size().await?;
            (i.0.zigzag(), i.1)
        })}
    }
    /// Read VarInt as i32 with size in bytes (varint, size)
    fn read_i32_varint_size(&mut self) -> impl Future<Output = Result<(i32, usize), ProtocolError>> {
        async {Ok({
            let i = self.read_u32_varint_size().await?;
            (i.0.zigzag(), i.1)
        })}
    }
    /// Read VarInt as i64 with size in bytes (varint, size)
    fn read_i64_varint_size(&mut self) -> impl Future<Output = Result<(i64, usize), ProtocolError>> {
        async {Ok({
            let i = self.read_u64_varint_size().await?;
            (i.0.zigzag(), i.1)
        })}
    }
    /// Read VarInt as i128 with size in bytes (varint, size)
    fn read_i128_varint_size(&mut self) -> impl Future<Output = Result<(i128, usize), ProtocolError>> {
        async {Ok({
            let i = self.read_u128_varint_size().await?;
            (i.0.zigzag(), i.1)
        })}
    }

    /// Read VarInt as usize
    fn read_usize_varint(&mut self) -> impl Future<Output = Result<usize, ProtocolError>> {
        async {read_varint!(usize, self)}
    }
    /// Read VarInt as u8
    fn read_u8_varint(&mut self) -> impl Future<Output = Result<u8, ProtocolError>> {
        async {read_varint!(u8, self)}
    }
    /// Read VarInt as u16
    fn read_u16_varint(&mut self) -> impl Future<Output = Result<u16, ProtocolError>> {
        async {read_varint!(u16, self)}
    }
    /// Read VarInt as u32
    fn read_u32_varint(&mut self) -> impl Future<Output = Result<u32, ProtocolError>> {
        async {read_varint!(u32, self)}
    }
    /// Read VarInt as u64
    fn read_u64_varint(&mut self) -> impl Future<Output = Result<u64, ProtocolError>> {
        async {read_varint!(u64, self)}
    }
    /// Read VarInt as u128
    fn read_u128_varint(&mut self) -> impl Future<Output = Result<u128, ProtocolError>> {
        async {read_varint!(u128, self)}
    }

    /// Read VarInt as isize
    fn read_isize_varint(&mut self) -> impl Future<Output = Result<isize, ProtocolError>> {
        async {Ok(self.read_usize_varint().await?.zigzag())}
    }
    /// Read VarInt as i8
    fn read_i8_varint(&mut self) -> impl Future<Output = Result<i8, ProtocolError>> {
        async {Ok(self.read_u8_varint().await?.zigzag())}
    }
    /// Read VarInt as i16
    fn read_i16_varint(&mut self) -> impl Future<Output = Result<i16, ProtocolError>> {
        async {Ok(self.read_u16_varint().await?.zigzag())}
    }
    /// Read VarInt as i32
    fn read_i32_varint(&mut self) -> impl Future<Output = Result<i32, ProtocolError>> {
        async {Ok(self.read_u32_varint().await?.zigzag())}
    }
    /// Read VarInt as i64
    fn read_i64_varint(&mut self) -> impl Future<Output = Result<i64, ProtocolError>> {
        async {Ok(self.read_u64_varint().await?.zigzag())}
    }
    /// Read VarInt as i128
    fn read_i128_varint(&mut self) -> impl Future<Output = Result<i128, ProtocolError>> {
        async {Ok(self.read_u128_varint().await?.zigzag())}
    }
}

impl<R: AsyncReadExt + Unpin> DataReader for R {
    async fn read_bytes(&mut self, size: usize) -> Result<Vec<u8>, ProtocolError> {
        let mut data = Vec::new();

        while data.len() < size {
            let mut buf = vec![0; size - data.len()];

            let n = self.read(&mut buf).await
                .map_err(|_| ProtocolError::ReadError)?;
            
            if n == 0 {
                return Err(ProtocolError::ConnectionClosedError);
            }
            
            buf.truncate(n);

            data.append(&mut buf);
        }

        Ok(data)
    }
}
